/*
** Copyright (C) 2002 Erik de Castro Lopo <erikd@mega-nerd.com>
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU Lesser General Public License as published by
** the Free Software Foundation; either version 2.1 of the License, or
** (at your option) any later version.
** 
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Lesser General Public License for more details.
** 
** You should have received a copy of the GNU Lesser General Public License
** along with this program; if not, write to the Free Software 
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include 	<pthread.h>
#include	<stdio.h>
#include	<unistd.h>
#include	<math.h>
#include	<complex.h>

#include	<sndfile.h>

#include	"gui.h"
#include	"common.h"
#include	"audio_out.h"

#define		BUFFER_SIZE		512

/* Milliseconds */
#define		MAX_DELAY		100.0  

#define	APPLY_CLIPPING(x,y)					\
			{ if ((x) > (y)) (x) = (y) ;	\
			  if ((x) < -(y)) (x) = -(y) ;	\
			  }

static void *gui_thread_proc (void*) ;
static void process_audio (SNDFILE *sndfile, int devdsp, int devmixer, int samplerate, int channels) ;

static char title [] = "Chorus/Flange" ;

static int	end_processing = 0 ;

static float	master_vol = 80.0, pcm_vol = 80.0 ;
static float	delay_time = MAX_DELAY, delay_mix = 0.0, feedback_amount = 0.0 ;
static float	mod_rate = 5.0, mod_depth = 0.0 ;

static GUI_ELEMENT elements [] = 
{	{	LINEAR_ADJUSTMENT,	"Master volume (%)", 
		{	0.0, 100.0, 0.1, NULL, &master_vol },	
		},
	{	LINEAR_ADJUSTMENT,	"PCM volume (%)", 
		{	0.0, 100.0, 0.1, NULL, &pcm_vol },	
		},
	{	HORIZ_SEPARATOR,	},
	{	LINEAR_ADJUSTMENT,	"Rate (Hz)", 
		{	0.1, 30.0, 0.1, NULL, &mod_rate },	
		},
	{	LINEAR_ADJUSTMENT,	"Depth (%)", 
		{	0.0, 20.0, 0.1, NULL, &mod_depth },	
		},
	{	HORIZ_SEPARATOR,	},
	{	LINEAR_ADJUSTMENT,	"Delay (ms)", 
		{	0.0, MAX_DELAY, 0.1, NULL, &delay_time },	
		},
	{	LINEAR_ADJUSTMENT,	"Feedback (%)", 
		{	0.0, 100.0, 0.1, NULL, &feedback_amount },	
		},
	{	LINEAR_ADJUSTMENT,	"Mix (%)", 
		{	0.0, 100.0, 0.1, NULL, &delay_mix },	
		},
} ;

int 
main (int argc, char *argv[])
{	SNDFILE  	*sndfile ;
	SF_INFO		sfinfo ;
	pthread_t	gui_thread ;
	int			devdsp, devmixer ;
	
	if (argc != 2)
		usage_exit (argv [0]) ;

	if (! (sndfile = sf_open_read (argv [1], &sfinfo)))
	{	printf ("Error : not able to open file '%s'\n", argv[1]) ;
		exit (1) ;
		} ;
	
	if (sfinfo.channels > 2)
	{	printf ("Error : file must be mono or stereo.\n") ;
		exit (1) ;
		} ;
		
	/* Always open as a stereo device. If we have a mono input signal, a
	** stereo output will be generated.
	*/
	devdsp = open_dev_dsp (2, sfinfo.samplerate) ;

	devmixer = open_dev_mixer () ;
		
	get_mixer (devmixer, &master_vol, &pcm_vol) ;

	if (pthread_create (&gui_thread, NULL, gui_thread_proc, NULL))
	{	printf ("Error creating screen thread.\n") ;	
		exit (1) ;
		} ;
	
	sleep (1) ; /* Pause for the GUI to start. */
	
	process_audio (sndfile, devdsp, devmixer, sfinfo.samplerate, sfinfo.channels) ;

	sf_close (sndfile) ;
	
	return 0 ;
} /* main */

static void *
gui_thread_proc (void* param)
{	
	/* This does not return until its time to kill program. */
	create_gui (title, elements, sizeof (elements) / sizeof (GUI_ELEMENT)) ;

	/* Signal all other threads that its time to die. */
	end_processing = 1 ;
	
	pthread_exit (NULL) ;
	
	return NULL ;
} /* gui_thread_proc */

typedef struct
{	float	*delay_l, *delay_r ;
	int		delay_len, index ;
} MOD_DELAY_DATA ;

static void
mod_delay_data_init (MOD_DELAY_DATA *mod_delay_data, int delay_len, int channels)
{		
	mod_delay_data->delay_l = malloc (channels * delay_len * sizeof (float)) ;
	mod_delay_data->delay_r = malloc (channels * delay_len * sizeof (float)) ;

	if (! mod_delay_data->delay_l || ! mod_delay_data->delay_l)
	{	printf ("Error : malloc failed.\n") ;
		exit (1) ;
		}

	memset (mod_delay_data->delay_l, 0, delay_len * sizeof (float)) ;
	memset (mod_delay_data->delay_r, 0, delay_len * sizeof (float)) ;

	mod_delay_data->index = 0 ;
	mod_delay_data->delay_len = delay_len ;
} /* mod_delay_data_init */

static void
mod_delay_data_fini (MOD_DELAY_DATA *mod_delay_data)
{	if (mod_delay_data->delay_l)
		free (mod_delay_data->delay_l) ;
	
	if (mod_delay_data->delay_r)
		free (mod_delay_data->delay_r) ;
	
	mod_delay_data->delay_l = NULL ;
	mod_delay_data->delay_r = NULL ;
} /* mod_delay_data_fini */

static void	apply_mod_delay_stereo (MOD_DELAY_DATA *mod_delay_data, short *buffer, int read_count, int samplerate) ;

static void duplicate_mono (short *buffer, int count) ;

static void 
process_audio (SNDFILE *sndfile, int devdsp, int devmixer, int samplerate, int channels)
{	static short buffer [BUFFER_SIZE] ;
	
	MOD_DELAY_DATA	mod_delay_data ;
	int read_count, delay_len ;
	
	/* Need two times the delay_len so thatwe can modulate beyond MAX_DELAY. */
	delay_len = ceil (2 * MAX_DELAY * samplerate / 1000.0) ;
	
	memset (&mod_delay_data, 0, sizeof (MOD_DELAY_DATA)) ;
	mod_delay_data_init (&mod_delay_data, delay_len, channels) ;
	
	while (! end_processing)
	{	set_mixer (devmixer, master_vol, pcm_vol) ;
		
		if (channels == 1)
		{	read_count = sf_read_short (sndfile, buffer, BUFFER_SIZE / 2) ;
			duplicate_mono (buffer, read_count) ;
			}
		else
			read_count = sf_read_short (sndfile, buffer, BUFFER_SIZE) ;
			
		if (read_count == 0)
		{	/* If we're at the end of the file, go back to start. */
			sf_seek (sndfile, 0, SEEK_SET) ;
			continue ;
			} ;
			
		apply_mod_delay_stereo (&mod_delay_data, buffer, read_count, samplerate) ;
		
		write (devdsp, buffer, read_count * sizeof (short)) ;
		} ;

	mod_delay_data_fini (&mod_delay_data) ;
} /* process_audio */

static void
apply_mod_delay_stereo (MOD_DELAY_DATA *mod_delay_data, short *buffer, int count, int samplerate)
{	static 	double complex	phasor = 1.0 ;	
	
	double 	sample, delay_sample, modulation ;
	int 	k, outdex ;
	
	/* Buffer must contais stereo data. */
	
	for (k = 0 ; k < count ; k += 2)
	{	phasor *= cexp (_Complex_I * 2.0 * M_PI * mod_rate / samplerate) ;
		phasor /= cabs (phasor) ;

		/*----------------------------------------------------------------
		** First do the left channel. 
		*/
		modulation = 1.0 + mod_depth * creal (phasor) / 100.0 ;
		
		/* Calculate the index of the sample coming out of the circular buffer. */ 
		outdex = mod_delay_data->index + mod_delay_data->delay_len - 
					floor (delay_time * modulation * samplerate / 1000.0) ;
		outdex %= mod_delay_data->delay_len ;

		/* Grab the sample from the calculated index. */
		delay_sample = mod_delay_data->delay_l [outdex] ;

		/* Calculate (and clip) the sample going into the circular buffer. */
		sample = buffer [k] + feedback_amount * delay_sample / 100.0 ;
		APPLY_CLIPPING (sample, 32767.0) ;
		mod_delay_data->delay_l [mod_delay_data->index] = sample ;
		
		/* Increment the current index modulo the length of the delay line. */
		mod_delay_data->index = (mod_delay_data->index + 1) % mod_delay_data->delay_len ;

		/* Mix the direct signal with the signal from the circular buffer. */
		sample = sqrt (1.0 - delay_mix / 100.0) * buffer [k] + sqrt (delay_mix / 100.0) * delay_sample ;
		APPLY_CLIPPING (sample, 32767.0) ;
		buffer [k] = sample ;

		/*----------------------------------------------------------------
		** Now do the right channel. 
		*/
		modulation = 1.0 + mod_depth * cimag (phasor) / 100.0 ;
		
		/* Calculate the index of the sample coming out of the circular buffer. */ 
		outdex = mod_delay_data->index + mod_delay_data->delay_len - 
					floor (delay_time * modulation * samplerate / 1000.0) ;
		outdex %= mod_delay_data->delay_len ;

		/* Grab the sample from the calculated index. */
		delay_sample = mod_delay_data->delay_r [outdex] ;

		/* Calculate (and clip) the sample going into the circular buffer. */
		sample = buffer [k + 1] + feedback_amount * delay_sample / 100.0 ;
		APPLY_CLIPPING (sample, 32767.0) ;
		mod_delay_data->delay_r [mod_delay_data->index] = sample ;
		
		/* Increment the current index modulo the length of the delay line. */
		mod_delay_data->index = (mod_delay_data->index + 1) % mod_delay_data->delay_len ;

		/* Mix the direct signal with the signal from the circular buffer. */
		sample = sqrt (1.0 - delay_mix / 100.0) * buffer [k + 1] + sqrt (delay_mix / 100.0) * delay_sample ;
		APPLY_CLIPPING (sample, 32767.0) ;
		buffer [k + 1] = sample ;
		} ;
} /* apply_mod_delay_stereo */


static void 
duplicate_mono (short *buffer, int count)
{	/* Convert a buffer of mono samples to pseudo stereo. 
	** This routine assumes that the buffer is twice as long as count,
	** the number of samples currently in the buffer. 
	*/ 

	while (count)
	{	count -- ;
	
		buffer [2 * count + 1] = buffer [count] ;
		buffer [2 * count    ] = buffer [count] ;
		} ;
} /* duplicate_mono */ 
