/*
** Copyright (C) 2002 Erik de Castro Lopo <erikd@mega-nerd.com>
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU Lesser General Public License as published by
** the Free Software Foundation; either version 2.1 of the License, or
** (at your option) any later version.
** 
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Lesser General Public License for more details.
** 
** You should have received a copy of the GNU Lesser General Public License
** along with this program; if not, write to the Free Software 
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include 	<pthread.h>
#include	<stdio.h>
#include	<unistd.h>
#include	<math.h>

#include	<sndfile.h>

#include	"gui.h"
#include	"common.h"
#include	"audio_out.h"

#define		BUFFER_SIZE		512

#define	APPLY_CLIPPING(x,y)					\
			{ if ((x) > (y)) (x) = (y) ;	\
			  if ((x) < -(y)) (x) = -(y) ;	\
			  }

#define		HARD_KNEE		0

static void *gui_thread_proc (void*) ;
static void process_audio (SNDFILE *sndfile, int devdsp, int devmixer, int samplerate, int channels) ;
static void	apply_comp (short *buffer, int read_count, int samplerate, int channels) ;

static char title [] = "Compression (Dynamic Range)" ;

static int	end_processing = 0 ;

static float	master_vol = 80.0, pcm_vol = 80.0 ;
static float	comp_attack = 5.0, comp_hold = 10.0, comp_release = 200.0 ;
static float	comp_ratio = 3.0, comp_threshold = -10.0, comp_outgain = 5.0 ;
static int		comp_type = 0 ;

static GUI_ELEMENT elements [] = 
{	{	LINEAR_ADJUSTMENT,	"Master volume (%)", 
		{	0.0, 100.0, 0.1, NULL, &master_vol },	
		},
	{	LINEAR_ADJUSTMENT,	"PCM volume (%)", 
		{	0.0, 100.0, 0.1, NULL, &pcm_vol },	
		},

	{	HORIZ_SEPARATOR,	},

	{	FLIPFLOP_CHOICE,	"Compressor Type", 
		{ },
		{ { "Feed-forward (Hard knee)", "Feedback (soft knee)", NULL }, NULL, &comp_type },
		},

	{	LINEAR_ADJUSTMENT,	"Attack (ms)", 
		{	0.1, 40.0, 0.1, NULL, &comp_attack },	
		},
	{	LINEAR_ADJUSTMENT,	"Hold (ms)", 
		{	0.1, 20.0, 0.1, NULL, &comp_hold },	
		},
	{	LINEAR_ADJUSTMENT,	"Release (ms)", 
		{	5.0, 2000.0, 0.1, NULL, &comp_release },	
		},
	{	LINEAR_ADJUSTMENT,	"Threshold (dB)", 
		{	-80.0, 0.0, 0.1, NULL, &comp_threshold },
		},
	{	LINEAR_ADJUSTMENT,	"Ratio (N:1)", 
		{	1.0, 30.0, 0.1, NULL, &comp_ratio },	
		},
	{	LINEAR_ADJUSTMENT,	"Makeup Gain (dB)", 
		{	1.0, 50.0, 0.1, NULL, &comp_outgain },	
		},
} ;

int 
main (int argc, char *argv[])
{	SNDFILE  	*sndfile ;
	SF_INFO		sfinfo ;
	pthread_t	gui_thread ;
	int			devdsp, devmixer ;
	
	if (argc != 2)
		usage_exit (argv [0]) ;

	if (! (sndfile = sf_open_read (argv [1], &sfinfo)))
	{	printf ("Error : not able to open file '%s'\n", argv [1]) ;
		exit (1) ;
		} ;
	
	if (sfinfo.channels > 2)
	{	printf ("Error : file must be mono or stereo.\n") ;
		exit (1) ;
		} ;
		
	devdsp = open_dev_dsp (sfinfo.channels, sfinfo.samplerate) ;

	devmixer = open_dev_mixer () ;
		
	get_mixer (devmixer, &master_vol, &pcm_vol) ;

	if (pthread_create (&gui_thread, NULL, gui_thread_proc, NULL))
	{	printf ("Error creating screen thread.\n") ;	
		exit (1) ;
		} ;
	
	sleep (1) ; /* Pause for the GUI to start. */
	
	process_audio (sndfile, devdsp, devmixer, sfinfo.samplerate, sfinfo.channels) ;

	sf_close (sndfile) ;
	
	return 0 ;
} /* main */

static void *
gui_thread_proc (void* param)
{	
	/* This does not return until its time to kill program. */
	create_gui (title, elements, sizeof (elements) / sizeof (GUI_ELEMENT)) ;

	/* Signal all other threads that its time to die. */
	end_processing = 1 ;
	
	pthread_exit (NULL) ;
	
	return NULL ;
} /* gui_thread_proc */

static void 
process_audio (SNDFILE *sndfile, int devdsp, int devmixer, int samplerate, int channels)
{	static short buffer [BUFFER_SIZE] ;

	int read_count ;
	
	while (! end_processing)
	{	set_mixer (devmixer, master_vol, pcm_vol) ;
		
		read_count = sf_read_short (sndfile, buffer, BUFFER_SIZE) ;

		if (read_count == 0)
		{	/* If we're at the end of the file, go back to start. */
			sf_seek (sndfile, 0, SEEK_SET) ;
			continue ;
			} ;
			
		apply_comp (buffer, read_count, samplerate, channels) ;
		
		write (devdsp, buffer, read_count * sizeof (short)) ;
		} ;

} /* process_audio */

/* This is a feed-forward compressor. */

enum
{	STATE_ATTACK,
	STATE_HOLD,
	STATE_RELEASE
} ;

static void
apply_comp (short *buffer, int read_count, int samplerate, int channels)
{	static double 	smoothed_peak = 0.0, current_gain = 1.0 ;
	static int		state = STATE_RELEASE, last_state = STATE_RELEASE, hold_count = 0 ;
	
	double	insta_peak = 0.0, lin_threshold, sample_0 = 0.0, sample_1 = 0.0 ;
	int 	k ;
	
	for (k = 0 ; k < read_count; k+= channels)
	{	if (comp_type != HARD_KNEE)
		{	sample_0 = pow (10.0, comp_outgain / 20.0) * current_gain * buffer [k] ;
			if (channels > 1)
			{	sample_1 = pow (10.0, comp_outgain / 20.0) * current_gain * buffer [k + 1] ;
				}
			} ;
	
		/* Find the max abs value of the current stereo pair. */
		insta_peak = fabs (buffer [k]) ;
		if (channels > 1 && fabs (buffer [k + 1]) > insta_peak)
			insta_peak = fabs (buffer [k + 1]) ;
		
		/* Normalize the peak value. */
		insta_peak /= 32768.0 ;
		
		/* Apply attack/hold/release times. */
		if (insta_peak > smoothed_peak)
		{	smoothed_peak = smoothed_peak + (insta_peak - smoothed_peak) *
								1000.0 * comp_attack / samplerate ;
			last_state = state ;
			state = STATE_ATTACK ;
			}
		else if (state == STATE_ATTACK)
		{	/* Previous state was attack, now move to hold state. */
			last_state = state ;
			state = STATE_HOLD ;
			hold_count = 1000.0 * comp_hold / samplerate ;
			}
		else if (state == STATE_HOLD && hold_count > 0)
		{	/*	Smoothed peak does not change during hold state. */
			hold_count -- ;
			}
		else
		{	last_state = state ;
			state = STATE_RELEASE ;
			/* Value of smoothed_peak decreases. */
			smoothed_peak = smoothed_peak - smoothed_peak *
								1000.0 * comp_attack / samplerate ;
			} ;
		
		/* Calculate the linear threshold value. */	
		lin_threshold = pow (10, comp_threshold / 20.0) ;

		/* Calculate current gain value. */		
		current_gain = 1.0 ;
		if (smoothed_peak > lin_threshold)
			current_gain = (lin_threshold + (smoothed_peak - lin_threshold) / comp_ratio) / smoothed_peak ;

		if (comp_type == HARD_KNEE)
		{	sample_0 = pow (10.0, comp_outgain / 20.0) * current_gain * buffer [k] ;
			if (channels > 1)
			{	sample_1 = pow (10.0, comp_outgain / 20.0) * current_gain * buffer [k + 1] ;
				}
			} ;

		APPLY_CLIPPING (sample_0, 32767.0) ;
		buffer [k] = sample_0 ;
		
		if (channels > 1)
		{	APPLY_CLIPPING (sample_1, 32767.0) ;
			buffer [k + 1] = sample_1 ;
			} ;
		
		} ;
		
} /* apply_comp */

