/*
** Copyright (C) 2002 Erik de Castro Lopo <erikd@mega-nerd.com>
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU Lesser General Public License as published by
** the Free Software Foundation; either version 2.1 of the License, or
** (at your option) any later version.
** 
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Lesser General Public License for more details.
** 
** You should have received a copy of the GNU Lesser General Public License
** along with this program; if not, write to the Free Software 
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include 	<pthread.h>
#include	<stdio.h>
#include	<unistd.h>
#include	<math.h>

#include	<sndfile.h>

#include	"gui.h"
#include	"common.h"
#include	"audio_out.h"

#define		BUFFER_SIZE		512

#define		MAX_DELAY		1000.0  /* Milliseconds */

#define	APPLY_CLIPPING(x,y)					\
			{ if ((x) > (y)) (x) = (y) ;	\
			  if ((x) < -(y)) (x) = -(y) ;	\
			  }

static void *gui_thread_proc (void*) ;
static void process_audio (SNDFILE *sndfile, int devdsp, int devmixer, int samplerate, int channels) ;

static char title [] = "Echo" ;

static int	end_processing = 0 ;

static float	master_vol = 80.0, pcm_vol = 80.0 ;
static float	echo_delay = 200.0, echo_mix = 0.0, echo_feedback = 0.0 ;

static GUI_ELEMENT elements [] = 
{	{	LINEAR_ADJUSTMENT,	"Master volume (%)", 
		{	0.0, 100.0, 0.1, NULL, &master_vol },	
		},
	{	LINEAR_ADJUSTMENT,	"PCM volume (%)", 
		{	0.0, 100.0, 0.1, NULL, &pcm_vol },	
		},
	{	HORIZ_SEPARATOR,	},
	{	LINEAR_ADJUSTMENT,	"Delay (ms)", 
		{	0.0, MAX_DELAY, 0.1, NULL, &echo_delay },	
		},
	{	LINEAR_ADJUSTMENT,	"Feedback (%)", 
		{	0.0, 100.0, 0.1, NULL, &echo_feedback },	
		},
	{	LINEAR_ADJUSTMENT,	"Mix (%)", 
		{	0.0, 100.0, 0.1, NULL, &echo_mix },	
		},
} ;

int 
main (int argc, char *argv[])
{	SNDFILE  	*sndfile ;
	SF_INFO		sfinfo ;
	pthread_t	gui_thread ;
	int			devdsp, devmixer ;
	
	if (argc != 2)
		usage_exit (argv [0]) ;

	if (! (sndfile = sf_open_read (argv [1], &sfinfo)))
	{	printf ("Error : not able to open file '%s'\n", argv[1]) ;
		exit (1) ;
		} ;
	
	if (sfinfo.channels > 2)
	{	printf ("Error : file must be mono or stereo.\n") ;
		exit (1) ;
		} ;
		
	devdsp = open_dev_dsp (sfinfo.channels, sfinfo.samplerate) ;

	devmixer = open_dev_mixer () ;
		
	get_mixer (devmixer, &master_vol, &pcm_vol) ;

	if (pthread_create (&gui_thread, NULL, gui_thread_proc, NULL))
	{	printf ("Error creating screen thread.\n") ;	
		exit (1) ;
		} ;
	
	sleep (1) ; /* Pause for the GUI to start. */
	
	process_audio (sndfile, devdsp, devmixer, sfinfo.samplerate, sfinfo.channels) ;

	sf_close (sndfile) ;
	
	return 0 ;
} /* main */

static void *
gui_thread_proc (void* param)
{	
	/* This does not return until its time to kill program. */
	create_gui (title, elements, sizeof (elements) / sizeof (GUI_ELEMENT)) ;

	/* Signal all other threads that its time to die. */
	end_processing = 1 ;
	
	pthread_exit (NULL) ;
	
	return NULL ;
} /* gui_thread_proc */

typedef struct
{	float	*delay ;
	int		delay_len, index ;
} ECHO_DATA ;

static void
echo_data_init (ECHO_DATA *echo_data, int delay_len, int channels)
{		
	echo_data->delay = malloc (channels * delay_len * sizeof (float)) ;

	if (! echo_data->delay)
	{	printf ("Error : malloc failed.\n") ;
		exit (1) ;
		}

	memset (echo_data->delay, 0, delay_len * sizeof (float)) ;

	echo_data->index = 0 ;
	echo_data->delay_len = delay_len ;
} /* echo_data_init */

static void
echo_data_fini (ECHO_DATA *echo_data)
{	if (echo_data->delay)
		free (echo_data->delay) ;
	
	echo_data->delay = NULL ;
} /* echo_data_fini */

static void	apply_echo (ECHO_DATA *echo_data, short *buffer, int read_count, int samplerate, int channels) ;

#define SILENT_START	200

static void 
process_audio (SNDFILE *sndfile, int devdsp, int devmixer, int samplerate, int channels)
{	static short buffer [BUFFER_SIZE] ;
	
	ECHO_DATA	echo_data ;
	int read_count, delay_len, silent_count = SILENT_START ;
	
	delay_len = ceil (MAX_DELAY * samplerate / 1000.0) ;
	
	memset (&echo_data, 0, sizeof (ECHO_DATA)) ;
	echo_data_init (&echo_data, delay_len, channels) ;
	
	while (! end_processing)
	{	set_mixer (devmixer, master_vol, pcm_vol) ;
		
		read_count = sf_read_short (sndfile, buffer, BUFFER_SIZE) ;

		if (read_count == 0)
		{	if (silent_count == 0)
			{	silent_count = SILENT_START ;

				/* If we're at the end of the file, go back to start. */
				sf_seek (sndfile, 0, SEEK_SET) ;
				continue ;
				}
			
			read_count = BUFFER_SIZE ;
			memset (buffer, 0, sizeof (buffer)) ;
			read_count = BUFFER_SIZE ;
			silent_count -- ;
			}
			
		apply_echo (&echo_data, buffer, read_count, samplerate, channels) ;
		
		write (devdsp, buffer, read_count * sizeof (short)) ;
		} ;

	echo_data_fini (&echo_data) ;
} /* process_audio */

static void
apply_echo (ECHO_DATA *echo_data, short *buffer, int count, int samplerate, int channels)
{	double 	sample, delay_sample ;
	int 	k, outdex ;
	
	/* Don't need to do anything complicated here for mono/stereo. */
	for (k = 0 ; k < count; k++)
	{	
	
		/* Calculate the index of the sample coming out of the circular buffer. */ 
		outdex = echo_data->index + echo_data->delay_len - floor (echo_delay * samplerate / 1000.0) ;
		outdex %= echo_data->delay_len ;
		
		/* Grab the sample from the calculated index. */
		delay_sample = echo_data->delay [outdex * channels] ;
		
		/* Calculate (and clip) the sample going into the circular buffer. */
		sample = buffer [k] + echo_feedback * delay_sample / 100.0 ;
		APPLY_CLIPPING (sample, 32767.0) ;
		echo_data->delay [echo_data->index * channels] = sample ;
		
		/* Increment the current index modulo the length of the delay line. */
		echo_data->index = (echo_data->index + 1) % echo_data->delay_len ;

		/* Mix the direct signal with the signal from the circular buffer. */
		sample = sqrt (1.0 - echo_mix / 100.0) * buffer [k] + sqrt (echo_mix / 100.0) * delay_sample ;
		APPLY_CLIPPING (sample, 32767.0) ;
		buffer [k] = sample ;
		} ;
} /* apply_echo */

