/*
** Copyright (C) 2002 Erik de Castro Lopo <erikd@mega-nerd.com>
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU Lesser General Public License as published by
** the Free Software Foundation; either version 2.1 of the License, or
** (at your option) any later version.
** 
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Lesser General Public License for more details.
** 
** You should have received a copy of the GNU Lesser General Public License
** along with this program; if not, write to the Free Software 
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include 	<pthread.h>
#include	<stdio.h>
#include	<unistd.h>
#include	<math.h>

#include	<sndfile.h>

#include	"gui.h"
#include	"common.h"
#include	"audio_out.h"

#define		BUFFER_SIZE		512

static void *gui_thread_proc (void*) ;
static void process_audio (SNDFILE *sndfile, int devdsp, int devmixer, int samplerate, int channels) ;
static void	apply_eq (short *buffer, int read_count, int samplerate, int channels) ;

static char title [] = "Parametric Equalizer" ;

static int	end_processing = 0 ;

static float	master_vol = 80.0, pcm_vol = 80.0 ;
static float	eq_freq = 400.0, eq_qfact = 10.0, eq_gain = 0.0 ;

static GUI_ELEMENT elements [] = 
{	{	LINEAR_ADJUSTMENT,	"Master volume (%)", 
		{	0.0, 100.0, 0.1, NULL, &master_vol },	
		},
	{	LINEAR_ADJUSTMENT,	"PCM volume (%)", 
		{	0.0, 100.0, 0.1, NULL, &pcm_vol },	
		},
	{	HORIZ_SEPARATOR,	},
	{	LOG_ADJUSTMENT,	"Frequency (Hz)", 
		{	20.0, 20000.0, 0.1, NULL, &eq_freq },	
		},
	{	LINEAR_ADJUSTMENT,	"Q Factor", 
		{	1.0, 100.0, 0.1, NULL, &eq_qfact },	
		},
	{	LINEAR_ADJUSTMENT,	"Gain", 
		{	-20.0, 20.0, 0.1, NULL, &eq_gain },	
		},
} ;

int 
main (int argc, char *argv[])
{	SNDFILE  	*sndfile ;
	SF_INFO		sfinfo ;
	pthread_t	gui_thread ;
	int			devdsp, devmixer ;
	
	if (argc != 2)
		usage_exit (argv [0]) ;

	if (! (sndfile = sf_open_read (argv [1], &sfinfo)))
	{	printf ("Error : not able to open file '%s'\n", argv [1]) ;
		exit (1) ;
		} ;
	
	if (sfinfo.channels > 2)
	{	printf ("Error : file must be mono or stereo.\n") ;
		exit (1) ;
		} ;
		
	devdsp = open_dev_dsp (sfinfo.channels, sfinfo.samplerate) ;

	devmixer = open_dev_mixer () ;
		
	get_mixer (devmixer, &master_vol, &pcm_vol) ;

	if (pthread_create (&gui_thread, NULL, gui_thread_proc, NULL))
	{	printf ("Error creating screen thread.\n") ;	
		exit (1) ;
		} ;
	
	sleep (1) ; /* Pause for the GUI to start. */
	
	process_audio (sndfile, devdsp, devmixer, sfinfo.samplerate, sfinfo.channels) ;

	sf_close (sndfile) ;
	
	return 0 ;
} /* main */

static void *
gui_thread_proc (void* param)
{	
	/* This does not return until its time to kill program. */
	create_gui (title, elements, sizeof (elements) / sizeof (GUI_ELEMENT)) ;

	/* Signal all other threads that its time to die. */
	end_processing = 1 ;
	
	pthread_exit (NULL) ;
	
	return NULL ;
} /* gui_thread_proc */

static void 
process_audio (SNDFILE *sndfile, int devdsp, int devmixer, int samplerate, int channels)
{	static short buffer [BUFFER_SIZE] ;

	int read_count ;
	
	while (! end_processing)
	{	set_mixer (devmixer, master_vol, pcm_vol) ;
		
		read_count = sf_read_short (sndfile, buffer, BUFFER_SIZE) ;

		if (read_count == 0)
		{	/* If we're at the end of the file, go back to start. */
			sf_seek (sndfile, 0, SEEK_SET) ;
			continue ;
			} ;
			
		apply_eq (buffer, read_count, samplerate, channels) ;
		
		write (devdsp, buffer, read_count * sizeof (short)) ;
		} ;

} /* process_audio */

static void
apply_eq (short *buffer, int read_count, int samplerate, int channels)
{	static double	x [2][2] = { { 0.0, 0.0 }, { 0.0, 0.0 } } ;
	static double	y [2][2] = { { 0.0, 0.0 }, { 0.0, 0.0 } } ;
	double	A, alpha, omega, cs, output ;
	double	a0, a1,a2,b0, b1, b2 ;
	int 	chan, k ;
	
/*-printf ("eq_freq = %f, eq_qfact = %f, eq_gain = %f\n", eq_freq, eq_qfact, eq_gain) ;-*/

	for (chan = 0 ; chan < channels ; chan ++)
		for (k = chan ; k < read_count; k+= channels)
		{	A     = pow (10.0, eq_gain / 20.0) ;
			omega = 2.0 * M_PI * eq_freq / samplerate ;
			alpha = sin (omega) / (2.0 * eq_qfact) ;
			cs    = cos (omega) ;
	
			b0 =   1 + alpha * A ;
			b1 =  -2 * cs ;
			b2 =   1 - alpha * A ;
			a0 =   1 + alpha / A ;
			a1 =  -2 * cs ;
			a2 =   1 - alpha / A ;
	
			/*
			y [n] = (b0/a0) * x [n] + (b1/a0) * x [n-1] + (b2/a0) * x [n-2] - (a1/a0) * 
						y [n-1] - (a2/a0) * y [n-2] ;
			*/
			
			output = (b0/a0) * buffer [k] + (b1/a0) * x [chan][0] + (b2/a0) * x [chan][1] - 
						(a1/a0) * y [chan][0] - (a2/a0) * y [chan][1] ;
	
			x [chan][1] = x [chan][0] ;
			x [chan][0] = buffer [k] ; 
	
			y [chan][1] = y [chan][0] ;
			y [chan][0] = output ;
			
			/* Clip signal to keep it in range or [MIN_SHORT..MAX_SHORT] */
			if (output > 32767)
				output = 32767 ;
			else if (output < -32767)
				output = -32767 ;

			buffer [k] = output ;
			} ;
		
	
} /* apply_eq */

