/*
** Copyright (C) 2002 Erik de Castro Lopo <erikd@mega-nerd.com>
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU Lesser General Public License as published by
** the Free Software Foundation; either version 2.1 of the License, or
** (at your option) any later version.
** 
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Lesser General Public License for more details.
** 
** You should have received a copy of the GNU Lesser General Public License
** along with this program; if not, write to the Free Software 
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include 	<pthread.h>
#include	<stdio.h>
#include	<unistd.h>
#include	<math.h>

#include	<sndfile.h>

#include	"gui.h"
#include	"common.h"
#include	"audio_out.h"

#define		BUFFER_SIZE		512

#define	APPLY_CLIPPING(x,y)					\
			{ if ((x) > (y)) (x) = (y) ;	\
			  if ((x) < -(y)) (x) = -(y) ;	\
			  }

#define		MAX_TAPS		5

static void *gui_thread_proc (void*) ;
static void process_audio (SNDFILE *sndfile, int devdsp, int devmixer, int samplerate, int channels) ;

static char title [] = "Reverb" ;

static int	end_processing = 0 ;

static float	master_vol = 80.0, pcm_vol = 80.0 ;
static float	reverb_delay = 40.0, reverb_feedback = 50.0, reverb_mix = 30.0 ;

static GUI_ELEMENT elements [] = 
{	{	LINEAR_ADJUSTMENT,	"Master volume (%)", 
		{	0.0, 100.0, 0.1, NULL, &master_vol },	
		},
	{	LINEAR_ADJUSTMENT,	"PCM volume (%)", 
		{	0.0, 100.0, 0.1, NULL, &pcm_vol },	
		},
	{	HORIZ_SEPARATOR,	},
	{	LINEAR_ADJUSTMENT,	"Delay (ms)", 
		{	30.0, 120.0, 0.1, NULL, &reverb_delay },	
		},
	{	LINEAR_ADJUSTMENT,	"Delay Feedback (%)", 
		{	0.0, 100.0, 0.1, NULL, &reverb_feedback },	
		},
	{	LINEAR_ADJUSTMENT,	"Mix (%)", 
		{	0.0, 100.0, 0.1, NULL, &reverb_mix },	
		},
} ;

int 
main (int argc, char *argv[])
{	SNDFILE  	*sndfile ;
	SF_INFO		sfinfo ;
	pthread_t	gui_thread ;
	int			devdsp, devmixer ;
	
	if (argc != 2)
		usage_exit (argv [0]) ;

	if (! (sndfile = sf_open_read (argv [1], &sfinfo)))
	{	printf ("Error : not able to open file '%s'\n", argv [1]) ;
		exit (1) ;
		} ;
	
	if (sfinfo.channels > 2)
	{	printf ("Error : file must be mono or stereo.\n") ;
		exit (1) ;
		} ;
		
	devdsp = open_dev_dsp (sfinfo.channels, sfinfo.samplerate) ;

	devmixer = open_dev_mixer () ;
		
	get_mixer (devmixer, &master_vol, &pcm_vol) ;

	if (pthread_create (&gui_thread, NULL, gui_thread_proc, NULL))
	{	printf ("Error creating screen thread.\n") ;	
		exit (1) ;
		} ;
	
	sleep (1) ; /* Pause for the GUI to start. */
	
	process_audio (sndfile, devdsp, devmixer, sfinfo.samplerate, sfinfo.channels) ;

	sf_close (sndfile) ;
	
	return 0 ;
} /* main */

static void *
gui_thread_proc (void* param)
{	
	/* This does not return until its time to kill program. */
	create_gui (title, elements, sizeof (elements) / sizeof (GUI_ELEMENT)) ;

	/* Signal all other threads that its time to die. */
	end_processing = 1 ;
	
	pthread_exit (NULL) ;
	
	return NULL ;
} /* gui_thread_proc */

#define MAX_PATHS			5
#define MAX_DELAY_LEN		10000

typedef struct
{	float 	delay [MAX_PATHS][MAX_DELAY_LEN] ;
	int		index ;
} REVERB_DATA ;

static void	apply_reverb (REVERB_DATA *reverb_data, short *buffer, int read_count, int samplerate, int channels) ;

#define SILENT_START	200

static void 
process_audio (SNDFILE *sndfile, int devdsp, int devmixer, int samplerate, int channels)
{	static short buffer [BUFFER_SIZE] ;
	
	REVERB_DATA	reverb_data ;
	int read_count, silent_count = SILENT_START ;
	
	memset (&reverb_data, 0, sizeof (REVERB_DATA)) ;
	
	while (! end_processing)
	{	set_mixer (devmixer, master_vol, pcm_vol) ;
		
		read_count = sf_read_short (sndfile, buffer, BUFFER_SIZE) ;

		if (read_count == 0)
		{	if (silent_count == 0)
			{	silent_count = SILENT_START ;

				/* If we're at the end of the file, go back to start. */
				sf_seek (sndfile, 0, SEEK_SET) ;
				continue ;
				}
			
			read_count = BUFFER_SIZE ;
			memset (buffer, 0, sizeof (buffer)) ;
			read_count = BUFFER_SIZE ;
			silent_count -- ;
			}
			
		apply_reverb (&reverb_data, buffer, read_count, samplerate, channels) ;
		
		write (devdsp, buffer, read_count * sizeof (short)) ;
		} ;

	return ;
} /* process_audio */

static void
apply_reverb (REVERB_DATA *reverb_data, short *buffer, int count, int samplerate, int channels)
{	double 	sample, delay_sample, extra_delay, rev_out ;
	int 	k, m, outdex, delay_count ;
	
	/* Don't need to do anything complicated here for mono/stereo. */
	for (k = 0 ; k < count; k++)
	{	rev_out = 0.0 ;
		extra_delay = 0.0 ;
		
		for (m = 0 ; m < MAX_PATHS ; m++)
		{	delay_count = floor (reverb_delay * samplerate / 1000.0 + extra_delay) ;
			outdex = reverb_data->index + MAX_DELAY_LEN - delay_count ;
			outdex %= MAX_DELAY_LEN ;

			delay_sample = reverb_feedback * reverb_data->delay [m][outdex]  / 100.0 ;
		
			sample = buffer [k] + delay_sample ;
			reverb_data->delay [m][reverb_data->index] = sample ;
		
			if (extra_delay > 0.0)
				extra_delay *= 0.8 ;
			else
				extra_delay = delay_count * 1.7123 ;

			rev_out += sample ;
			} ;
			
		/* Increment the current index modulo the length of the delay line. */
		reverb_data->index = (reverb_data->index + 1) % MAX_DELAY_LEN ;

		sample = sqrt (1.0 - reverb_mix / 100.0) * buffer [k] + sqrt (reverb_mix / 100.0) * rev_out ;
		APPLY_CLIPPING (sample, 32767.0) ;
		buffer [k] = sample ;
		} ;
} /* apply_reverb */

