/*
** Copyright (C) 2002 Erik de Castro Lopo <erikd@mega-nerd.com>
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU Lesser General Public License as published by
** the Free Software Foundation; either version 2.1 of the License, or
** (at your option) any later version.
** 
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Lesser General Public License for more details.
** 
** You should have received a copy of the GNU Lesser General Public License
** along with this program; if not, write to the Free Software 
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include 	<pthread.h>
#include	<stdio.h>
#include	<unistd.h>
#include	<math.h>
#include	<complex.h>

#include	<sndfile.h>

#include	"gui.h"
#include	"common.h"
#include	"audio_out.h"

#define		BUFFER_SIZE		512

#define		MAX_DEPTH		50.0

static void *gui_thread_proc (void*) ;
static void process_audio (SNDFILE *sndfile, int devdsp, int devmixer, int samplerate, int channels) ;
static void	apply_tremelo (short *buffer, int read_count, int samplerate, int channels) ;
static double volume_warp (double depth, double x) ;

static char title [] = "Tremelo (Amplitude Modulation)" ;

static int	end_processing = 0 ;

static float	master_vol = 80.0, pcm_vol = 80.0 ;
static float	tremelo_rate = 5.0, tremelo_depth = 10.0 ;

static GUI_ELEMENT elements [] = 
{	{	LINEAR_ADJUSTMENT,	"Master volume (%)", 
		{	0.0, 100.0, 0.1, NULL, &master_vol },	
		},
	{	LINEAR_ADJUSTMENT,	"PCM volume (%)", 
		{	0.0, 100.0, 0.1, NULL, &pcm_vol },	
		},
	{	HORIZ_SEPARATOR,	},
	{	LINEAR_ADJUSTMENT,	"Rate (Hz)", 
		{	1.0, 30.0, 0.1, NULL, &tremelo_rate },	
		},
	{	LINEAR_ADJUSTMENT,	"Depth (%)", 
		{	0.0, MAX_DEPTH, 0.1, NULL, &tremelo_depth },	
		},
} ;

int 
main (int argc, char *argv[])
{	SNDFILE  	*sndfile ;
	SF_INFO		sfinfo ;
	pthread_t	gui_thread ;
	int			devdsp, devmixer ;
	
	if (argc != 2)
		usage_exit (argv[0]) ;

	if (! (sndfile = sf_open_read (argv [1], &sfinfo)))
	{	printf ("Error : not able to open file '%s'\n", argv [1]) ;
		exit (1) ;
		} ;
	
	if (sfinfo.channels > 2)
	{	printf ("Error : file must be mono or stereo.\n") ;
		exit (1) ;
		} ;
		
	devdsp = open_dev_dsp (sfinfo.channels, sfinfo.samplerate) ;

	devmixer = open_dev_mixer () ;
		
	get_mixer (devmixer, &master_vol, &pcm_vol) ;

	if (pthread_create (&gui_thread, NULL, gui_thread_proc, NULL))
	{	printf ("Error creating screen thread.\n") ;	
		exit (1) ;
		} ;
	
	sleep (1) ; /* Pause for the GUI to start. */
	
	process_audio (sndfile, devdsp, devmixer, sfinfo.samplerate, sfinfo.channels) ;

	sf_close (sndfile) ;
	
	return 0 ;
} /* main */

static void *
gui_thread_proc (void* param)
{	
	/* This does not return until its time to kill program. */
	create_gui (title, elements, sizeof (elements) / sizeof (GUI_ELEMENT)) ;

	/* Signal all other threads that its time to die. */
	end_processing = 1 ;
	
	pthread_exit (NULL) ;
	
	return NULL ;
} /* gui_thread_proc */

static void 
process_audio (SNDFILE *sndfile, int devdsp, int devmixer, int samplerate, int channels)
{	static short buffer [BUFFER_SIZE] ;

	int read_count ;
	
	while (! end_processing)
	{	set_mixer (devmixer, master_vol, pcm_vol) ;
		
		read_count = sf_read_short (sndfile, buffer, BUFFER_SIZE) ;

		if (read_count == 0)
		{	/* If we're at the end of the file, go back to start. */
			sf_seek (sndfile, 0, SEEK_SET) ;
			continue ;
			} ;
			
		apply_tremelo (buffer, read_count, samplerate, channels) ;
		
		write (devdsp, buffer, read_count * sizeof (short)) ;

		} ;

} /* process_audio */

static void
apply_tremelo (short *buffer, int read_count, int samplerate, int channels)
{	static 	double complex	phasor = 1.0 ;	
	int 	k ;
	
	for (k = 0 ; k < read_count; k++)
	{	/* Only update phasor on on the zeroth channel. */
		if ((k % channels) == 0)
		{	phasor *= cexp (_Complex_I * 2.0 * M_PI * tremelo_rate / samplerate) ;
			phasor /= cabs (phasor) ;
			} ;

		buffer [k] = buffer [k] * (volume_warp (tremelo_depth, cimag (phasor)) / 100.0) ;
		} ;
} /* apply_tremelo */

static double 
volume_warp (double depth, double x)
{	depth = 1.0 + depth / 4.0 ;

	x =  log10 (2.0 * MAX_DEPTH) - log10 (depth) + log10 (depth) * x ;

	return pow (10.0,x) ;
} /* volume_warp */
