## Copyright (C) 1995, 1998 Erik de Castro Lopo
##
## This file is part of Octave.
##
## Octave is free software; you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2, or (at your option)
## any later version.
##
## Octave is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
## General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with Octave; see the file COPYING.  If not, write to the Free
## Software Foundation, 59 Temple Place - Suite 330, Boston, MA
## 02111-1307, USA.

## Generate a linear phase FIR filter to fit a continuous piecewise linear 
## desired frequency response. The filter will have N+1 coefficients.
##        h = fircpl (N, freq, mag) ;

## Author: Erik de Castro Lopo <erikd@zip.com.au>
## Keywords: FIR, filter, remez

## Reference: Mike V. Chan et al.
##            "Formulas for the Impulse Response of a Digital Filter
##             with an Arbitrary Piecewise-Linear Frequency Response"
##             IEEE Trans. on Signal Processing
##             Vol. 43, No 1, January 1995, pp308-310

## Version: 1.3	 Fixed offset bug for even length filters.
##          1.2  Fixed indexing bug for non-integer N.
##          1.1	 Added checking for monotonically increasing freq. vector.

## Example: octave:1> freq = [0 0.2 0.4 0.6 0.8 1] ;
##          octave:2> mag = [1 1 0.5 0.5 0 0] ;
##          octave:3> b = fircpl (100, freq, mag);
##          octave:4> [h w] = freqz (b, 1, 512);
##          octave:5> plot (freq, mag, w/pi, abs (h))

function h = fircpl (N, freq, mag) ;

if nargin != 3
	error ("Three input arguments required.") ;
	endif
	
if max (size (N)) > 1,
	error ("Expecting a scalar for N.") ;
	endif
	
flen = length (freq) ;

if flen != length (mag),
    error ("Frequency and magnitude vectors must be the same length.") 
    endif ;

if freq (1) != 0,
    error ("First element in frequency vector must be zero.") 
    endif ;
    
if freq (flen) != 1,
    error ("Last element in frequency vector must be 1.") 
    endif ;
    
if (max (diff (freq) <= 0) > 0)
    error ('Frequency vector must be monotonically increasing.') 
    endif ;

Wsum  = (freq (1:flen-1) + freq (2:flen)) * pi ;
Wdiff = diff (freq) * pi ;

Msum  = mag (1:flen-1) + mag (2:flen) ;
Mdiff = diff (mag) ;

# By DSP convention, FIR filters have N+1 coefficients.

h = zeros (1, N+1);

# This is a direct implementation of equations 18 and 19 
# of the paper referenced above.

index = 1 ;
for n = -N/2:N/2,
	if n == 0,
		h (index) = 1 / 2 / pi * sum (Msum .* Wdiff) ;
	else
		h (index) = 2 / (n * n * pi) * ...
			sum (Mdiff ./ Wdiff .* sin (Wsum * n / 2) .* sin (-1 * Wdiff * n / 2)) ;
		endif 
	index ++ ;
	endfor

# Weight filter coefficients with hamming window before returning.

h = h .* (0.54 - 0.46 * cos (2 * pi * (0:N) / N)) ;

endfunction
